from typing import List, Tuple, Dict, Any, Set
import csv
import os
import sys

# Conditional import for TOML parsing
if sys.version_info >= (3, 11):
    import tomllib  # pylint: disable=unused-import
else:
    try:
        import tomli as tomllib  # Alias tomli as tomllib for consistent use
    except ImportError:
        tomllib = None  # type: ModuleType | None


def _load_toml_data(toml_file_path: str) -> Dict[str, Any]:
    """
    Loads data from a TOML file using the appropriate library (tomllib or tomli).
    Raises RuntimeError if tomli is needed but not installed on Python < 3.11.
    """
    if tomllib:
        with open(
            toml_file_path, "rb"
        ) as f:  # tomllib.load or tomli.load expects a binary file
            return tomllib.load(f)
    else:
        # This case should only be hit on Python < 3.11 if tomli failed to import
        raise RuntimeError(
            f"TOML host file ('{os.path.basename(toml_file_path)}') requires "
            "'tomli' to be installed on Python < 3.11. It should have been "
            "installed automatically with Ananta if using Python < 3.11. "
            "If not, please try reinstalling Ananta: "
            "pip install ananta --force-reinstall"
        )


def _get_hosts_from_toml(
    toml_file_path: str, host_tags_filter_str: str | None
) -> Tuple[List[Tuple[str, str, int, str, str]], int]:
    """
    Reads hosts from a TOML file and returns a list of tuples with host details.
    """
    hosts_to_execute: List[Tuple[str, str, int, str, str]] = []
    active_tags_filter: Set[str] = (
        set(host_tags_filter_str.split(",")) if host_tags_filter_str else set()
    )

    try:
        data = _load_toml_data(toml_file_path)
    except FileNotFoundError:
        print(f"Error: TOML hosts file not found at '{toml_file_path}'")
        return [], 0
    except RuntimeError as e:
        print(f"Error: {e}")
        return [], 0
    except tomllib.TOMLDecodeError if tomllib else Exception as e:
        print(f"Error decoding TOML file '{toml_file_path}': {e}")
        return [], 0
    except Exception as e:
        print(
            f"An unexpected error occurred while loading TOML file "
            f"'{toml_file_path}': {e}"
        )
        return [], 0

    defaults: Dict[str, Any] = data.get("default", {})
    default_port: int = defaults.get("port", 22)
    default_username: str | None = defaults.get("username")
    default_key_path: str = defaults.get("key_path", "#")
    # default tags won't be overridden but got appended by host-specific tags
    default_tags: List[str] = defaults.get("tags", [])

    for host_name, host_config in data.items():
        if host_name == "default":
            continue
        if not isinstance(host_config, dict):
            print(
                f"Warning: Skipping non-dictionary section '{host_name}' in "
                f"TOML file '{toml_file_path}'."
            )
            continue

        ip_address: str | None = host_config.get("ip")
        # ip_address accepts both IP address and resolvable hostname
        if not ip_address or not isinstance(ip_address, str):
            print(
                f"Warning: Host '{host_name}' in '{toml_file_path}' is missing "
                "'ip' or 'ip' is not a string. Skipping!"
            )
            continue

        try:
            port_str = host_config.get("port", default_port)
            ssh_port = int(port_str)
            username = host_config.get("username", default_username)
            if not username or not isinstance(username, str):
                print(
                    f"Warning: Host '{host_name}' in '{toml_file_path}' is missing "
                    "'username' or 'username' is not a string. Skipping!"
                )
                continue
            key_path = str(host_config.get("key_path", default_key_path))
            current_host_tags_list: List[str] = host_config.get("tags", [])
            if not isinstance(current_host_tags_list, list) or not all(
                isinstance(tag, str) for tag in current_host_tags_list
            ):
                print(
                    f"Warning: Host '{host_name}' in '{toml_file_path}' has "
                    "invalid 'tags' (must be a list of strings). Treating as no tags."
                )
                current_host_tags_set: Set[str] = set()
            else:
                # Append default tags to host-specific tags
                current_host_tags_set = set(
                    default_tags + current_host_tags_list
                )
            if not active_tags_filter or active_tags_filter.intersection(
                current_host_tags_set
            ):
                hosts_to_execute.append(
                    (host_name, ip_address, ssh_port, username, key_path)
                )
        except ValueError:
            print(
                f"Hosts file (TOML) '{toml_file_path}': Error parsing port for "
                f"host '{host_name}'. Port must be an integer. Skipping!"
            )
        except Exception as e:
            print(
                f"Hosts file (TOML) '{toml_file_path}': Unexpected error processing "
                f"host '{host_name}': {e}. Skipping!"
            )

    if hosts_to_execute:
        max_name_length = max(len(name) for name, *_ in hosts_to_execute)
        return hosts_to_execute, max_name_length
    return [], 0


def _get_hosts_from_csv(
    csv_file_path: str, host_tags_filter_str: str | None
) -> Tuple[List[Tuple[str, str, int, str, str]], int]:
    """
    Reads hosts from a CSV file and returns a list of tuples with host details.
    """
    hosts_to_execute: List[Tuple[str, str, int, str, str]] = []
    active_tags_filter: Set[str] = (
        set(host_tags_filter_str.split(",")) if host_tags_filter_str else set()
    )

    try:
        with open(csv_file_path, "r", encoding="utf-8") as hosts_file_obj:
            csv_reader = csv.reader(hosts_file_obj)
            for row_line, row in enumerate(csv_reader, start=1):
                if not row or row[0].startswith("#"):
                    continue

                # Check for minimum number of columns before unpacking
                if len(row) < 4:
                    print(
                        f"Hosts file (CSV): '{csv_file_path}' row {row_line} is incomplete "
                        "(expected at least 4 columns for name, ip, port, user). Skipping!"
                    )
                    continue

                try:
                    host_name, ip_address, str_port, username = row[:4]
                    ssh_port = int(
                        str_port
                    )  # ValueError here is specific to port format

                    key_path = row[4] if len(row) > 4 else ""
                    tags_in_csv_str = row[5] if len(row) > 5 else ""
                except ValueError:  # Catches error from int(str_port)
                    print(
                        f"Hosts file (CSV): '{csv_file_path}' parse error at row {row_line} "
                        "(port must be an integer). Skipping!"
                    )
                    continue
                # IndexError for key_path or tags_in_csv_str is avoided by conditional access

                current_host_tags_set: Set[str] = (
                    set(tags_in_csv_str.split(":"))
                    if tags_in_csv_str
                    else set()
                )
                if not active_tags_filter or active_tags_filter.intersection(
                    current_host_tags_set
                ):
                    hosts_to_execute.append(
                        (host_name, ip_address, ssh_port, username, key_path)
                    )
    except FileNotFoundError:
        print(f"Error: CSV hosts file not found at '{csv_file_path}'")
        return [], 0
    except Exception as e:
        print(
            f"An unexpected error occurred while reading CSV file '{csv_file_path}': {e}"
        )
        return [], 0

    if hosts_to_execute:
        max_name_length = max(len(name) for name, *_ in hosts_to_execute)
        return hosts_to_execute, max_name_length
    return [], 0


def get_hosts(
    host_file_path: str, host_tags: str | None
) -> Tuple[List[Tuple[str, str, int, str, str]], int]:
    """
    Reads hosts from a file (TOML or CSV) and returns a list of tuples with host details.
    """
    if not host_file_path:
        return [], 0
    _root, file_ext = os.path.splitext(host_file_path.lower())
    if file_ext == ".toml":
        return _get_hosts_from_toml(host_file_path, host_tags)
    elif file_ext == ".csv":
        return _get_hosts_from_csv(host_file_path, host_tags)
    else:
        print(
            f"Warning: Unknown or missing host file extension for "
            f"'{os.path.basename(host_file_path)}'. Attempting to parse as CSV."
        )
        return _get_hosts_from_csv(host_file_path, host_tags)



#!/usr/bin/env python
"""
Urwid-based Text User Interface for Ananta.
Manages asynchronous SSH connections and command execution on multiple remote hosts.
"""
from __future__ import annotations
from ..config import get_hosts
from ..ssh import establish_ssh_connection, stream_command_output
from .ansi import ansi_to_urwid_markup
from itertools import cycle
from random import shuffle
from typing import Any, Dict, List, Set, Tuple
import asyncio
import asyncssh
import re
import sys
import urwid


class ListBoxWithScrollBar(urwid.WidgetWrap):
    """A ListBox with a visual scrollbar."""

    def __init__(self, walker: urwid.SimpleFocusListWalker):
        self._walker = walker
        self._list_box = urwid.ListBox(self._walker)
        self._scrollbar = urwid.Text("", align="left")
        self._wrapped_widget = urwid.Columns(
            [
                (self._list_box),
                ("fixed", 1, urwid.AttrMap(self._scrollbar, "body")),
            ],
            box_columns=[0],
        )
        super().__init__(self._wrapped_widget)

    def render(
        self, size: Tuple[int, int], focus: bool = False
    ) -> urwid.Canvas:
        """Render the widget and update the scrollbar."""
        self._update_scrollbar(size)
        return super().render(size, focus)

    def _update_scrollbar(self, size: Tuple[int, int]) -> None:
        """Update the scrollbar's appearance based on the list box's state."""
        max_height = size[1]
        if max_height <= 0 or not self._walker:
            self._scrollbar.set_text("")
            return

        content_length = len(self._walker)
        if content_length <= max_height:
            self._scrollbar.set_text("")  # No scrollbar needed
            return

        focus_pos = self._walker.focus
        if not 0 <= focus_pos < content_length:
            focus_pos = content_length - 1

        if content_length > 1:
            scroll_ratio = focus_pos / (content_length - 1)
        else:
            scroll_ratio = 0

        handle_size = max(1, round(max_height * (max_height / content_length)))
        handle_size = min(max_height, handle_size)

        scrollable_space = max_height - handle_size
        handle_top = round(scroll_ratio * scrollable_space)

        bar_chars = []
        for i in range(max_height):
            if handle_top <= i < handle_top + handle_size:
                bar_chars.append("█")
            else:
                bar_chars.append("░")

        self._scrollbar.set_text("\n".join(bar_chars))

    def keypress(self, size: Tuple[int, int], key: str) -> str | None:
        """Pass keypresses to the list box."""
        if key == "mouse press":
            return key
        return self._list_box.keypress(size, key)

    def mouse_event(
        self,
        size: Tuple[int, int],
        event: str,
        button: int,
        col: int,
        row: int,
        focus: bool,
    ) -> bool | None:
        """Handle mouse events, specifically for scrolling."""
        if event == "mouse press":
            if button == 4:  # Scroll up
                self._list_box.keypress(size, "page up")
                return True
            if button == 5:  # Scroll down
                self._list_box.keypress(size, "page down")
                return True
        return self._list_box.mouse_event(size, event, button, col, row, focus)

    @property
    def body(self) -> urwid.SimpleFocusListWalker:
        """Provide access to the walker for external manipulation."""
        return self._walker


# --- Setup colors for hosts ---
URWID_FG_COLORS = [
    "yellow",
    "light red",
    "light green",
    "light blue",
    "light magenta",
    "light cyan",
]  # dark colors are not used to avoid confusion with similarity of colors
shuffle(URWID_FG_COLORS)  # Shuffle to randomize color assignment
COLORS_CYCLE = cycle(URWID_FG_COLORS)  # Create a cycle in case of many hosts


def _get_host_attr_name(host_name: str) -> str:
    return f"host_{host_name.lower().replace('-', '_').replace(' ', '_').replace('.', '_')}"


# format_host_prompt no longer needs palette_list or self
def format_host_prompt(
    host_name: str, max_name_length: int
) -> List[Tuple[str, str]]:
    attr_name = _get_host_attr_name(host_name)
    padded_host = host_name.rjust(max_name_length)
    return [(attr_name, f"[{padded_host}] ")]


class AnantaMainLoop(urwid.MainLoop):
    def entering_idle(self) -> None:
        """
        Override the base method to prevent automatic screen redraws on idle.
        This helps avoid `BlockingIOError` by giving us manual control over the redraw cycle.
        """
        pass


class RefreshingPile(urwid.Pile):
    """A Pile that forces a redraw after any keypress is handled."""

    def __init__(
        self, widget_list: List[Any], tui: "AnantaUrwidTUI", **kwargs: Any
    ):
        self._tui = tui
        super().__init__(widget_list, **kwargs)

    def keypress(self, size: Tuple[int, int], key: str) -> str | None:
        """Handle keypress and then request a redraw."""
        result = super().keypress(size, key)
        # After any keypress, handled or not by a child, request a redraw.
        # This is necessary because the main loop's idle handler is disabled.
        if (
            self._tui.loop
            and self._tui.loop.event_loop
            and not self._tui.draw_screen_handle
        ):
            self._tui.draw_screen_handle = self._tui.loop.event_loop.alarm(
                0, self._tui._request_draw
            )
        return result


class AnantaUrwidTUI:
    """Ananta Text User Interface using Urwid."""

    DEFAULT_PALETTE = [
        ("status_ok", "light green", "default", None, None, None),
        ("status_error", "light red", "default", None, None, None),
        ("status_neutral", "yellow", "default", None, None, None),
        ("command_echo", "light cyan,bold", "default", None, None, None),
        ("body", "white", "default", None, None, None),
        ("input_prompt", "light blue", "default", None, None, None),
        ("input_prompt_inactive", "dark gray", "default", None, None, None),
        ("ansi_bold", "bold", "default", None, None, None),
        ("ansi_underline", "underline", "default", None, None, None),
        ("ansi_standout", "standout", "default", None, None, None),
    ]

    def __init__(
        self,
        host_file: str,
        initial_command: str | None,
        host_tags: str | None,
        default_key: str | None,
        separate_output: bool,
        allow_empty_line: bool,
    ):
        """Initialize the Ananta TUI."""
        # --- SSH, connection, and output setup ---
        self.host_file = host_file
        self.initial_command = initial_command
        self.host_tags = host_tags
        self.default_key = default_key
        self.separate_output = separate_output
        self.allow_empty_line = allow_empty_line
        self.hosts, self.max_name_length = get_hosts(host_file, host_tags)
        self.connections: Dict[str, asyncssh.SSHClientConnection | None] = {
            host[0]: None for host in self.hosts
        }
        self.output_queues: Dict[str, asyncio.Queue[str | None]] = {
            host[0]: asyncio.Queue() for host in self.hosts
        }
        # --- Urwid setup ---
        self.host_palette_definitions: Dict[
            str, Tuple[str, str, str, None, None, None]
        ] = {}
        self._populate_host_palette_definitions()
        self.current_palette = self._build_palette()
        self.output_walker = urwid.SimpleFocusListWalker([])
        self.output_box = ListBoxWithScrollBar(self.output_walker)
        self.input_field = urwid.Edit(edit_text="")
        self.prompt_widget = urwid.Text(">>> ")
        self.prompt_attr_map = urwid.AttrMap(self.prompt_widget, "input_prompt")
        self.input_wrapper = urwid.Columns(
            [
                ("fixed", 4, self.prompt_attr_map),
                self.input_field,
            ],
            dividechars=0,
        )
        widgets = [
            ("weight", 1, urwid.AttrMap(self.output_box, "body")),
            ("fixed", 1, urwid.SolidFill("─")),
            ("fixed", 1, urwid.AttrMap(self.input_wrapper, "body")),
        ]
        self.main_pile = RefreshingPile(widgets, tui=self, focus_item=2)
        self.main_layout = urwid.Frame(body=self.main_pile)
        self.main_pile.focus_position = 2
        urwid.connect_signal(
            self.input_field, "change", self.update_prompt_attribute
        )
        # --- Event loop and async tasks setup ---
        self.loop: urwid.MainLoop | None = None
        self.async_tasks: Set[asyncio.Task[Any]] = set()
        self.is_exiting = False
        self.asyncio_loop: asyncio.AbstractEventLoop | None = None
        self.draw_screen_handle: Any = None
        self.shutdown_task: asyncio.Task[None] | None = None

        # --- Show the TUI welcome message ---
        self.add_output(
            [
                (
                    "status_ok",
                    "+------------------------------------------+\n"
                    "|        Welcome to Ananta TUI mode.       |\n"
                    "|Press [Up] to focus on the Output window. |\n"
                    "|Press [PgUp] or [PgDn] to scroll up/down. |\n"
                    "|Press [Down] to focus on the Input window.|\n"
                    "|Press [Ctrl-D] or `exit` command to exit. |\n"
                    "+------------------------------------------+",
                )
            ]
        )

        # --- Show warning for Separate Output mode ---
        if self.separate_output:
            self.add_output(
                [
                    (
                        "status_error",
                        "+------------------------------------------+\n"
                        "|    Separate Output mode [-s] enabled.    |\n"
                        "|    Avoid feeding LARGE amount of data.   |\n"
                        "|       !!!!You have been WARNED!!!!       |\n"
                        "+------------------------------------------+",
                    )
                ]
            )

    def _populate_host_palette_definitions(self) -> None:
        """Pre-populates host-specific palette entries."""
        for host_name, *_ in self.hosts:
            attr_name = _get_host_attr_name(host_name)
            if attr_name not in self.host_palette_definitions:
                fg_color = next(COLORS_CYCLE)
                self.host_palette_definitions[attr_name] = (
                    attr_name,
                    fg_color,
                    "default",
                    None,
                    None,
                    None,
                )

    def _build_palette(self) -> List[Tuple[str | None, ...]]:
        """Build the complete palette for Urwid, including default and host-specific styles."""
        palette = list(self.DEFAULT_PALETTE)

        # Add pre-defined host styles
        for host_entry in self.host_palette_definitions.values():
            # Simple check to avoid adding if somehow already present by name
            # More robust deduplication happens next anyway
            if not any(
                entry[0] == host_entry[0]
                for entry in palette
                if entry and entry[0]
            ):
                palette.append(host_entry)

        seen_names = set()
        unique_palette: List[Tuple[str | None, ...]] = []
        for entry in reversed(
            palette
        ):  # Iterate from the end to keep last definition of a name
            if isinstance(entry, tuple) and entry[0] is not None:
                if entry[0] not in seen_names:
                    unique_palette.insert(0, entry)
                    seen_names.add(entry[0])
        return unique_palette

    def add_output(
        self, message_parts: List[Any] | str, scroll: bool = True
    ) -> None:
        """Add output to the display."""
        if self.is_exiting and not any(
            s in str(message_parts).lower()
            for s in [
                "exiting",
                "closed",
                "cleanup",
                "shutdown",
                "processed",
                "failed",
                "error",
            ]
        ):
            return

        if isinstance(message_parts, str):
            processed_markup = ansi_to_urwid_markup(message_parts)
        else:
            processed_markup = message_parts

        if (
            not processed_markup
            and isinstance(message_parts, str)
            and message_parts.strip() == ""
        ):
            widget = urwid.Text("")
        elif not processed_markup:
            return
        else:
            widget = urwid.Text(processed_markup)

        self.output_walker.append(widget)

        rows: int = 24
        if self.loop and self.loop.screen:
            _, rows = self.loop.screen.get_cols_rows()
        max_lines = rows * 10
        trim_lines = rows
        if len(self.output_walker) > max_lines:
            del self.output_walker[
                0 : len(self.output_walker) - (max_lines - trim_lines)
            ]

        if scroll:
            self.output_walker.set_focus(len(self.output_walker) - 1)

        if self.loop and self.loop.event_loop and not self.draw_screen_handle:
            self.draw_screen_handle = self.loop.event_loop.alarm(
                0, self._request_draw
            )

    def _request_draw(self, *_args: Any) -> None:
        """Request a redraw of the screen."""
        try:
            if self.loop:
                self.loop.draw_screen()
        except BlockingIOError:
            pass  # Ignore if the screen is busy
        finally:
            self.draw_screen_handle = None

    async def connect_host(
        self,
        host_name: str,
        ip: str,
        port: int,
        user: str,
        key: str,
    ) -> None:
        """Establish an SSH connection to a single host."""
        if self.is_exiting:
            return

        prompt = format_host_prompt(
            host_name, self.max_name_length
        )  # No longer needs self.current_palette
        self.add_output(prompt + [("status_neutral", "Connecting...")])

        try:
            conn = await establish_ssh_connection(
                ip, port, user, key, self.default_key, timeout=10.0
            )
        except Exception as e:
            self.connections[host_name] = None
            self.add_output(
                prompt + [("status_error", f"Connection failed: {e}")]
            )
        else:
            conn.set_keepalive(interval=30, count_max=3)
            self.connections[host_name] = conn
            self.add_output(prompt + [("status_ok", "Connected.")])

    async def connect_all_hosts(self) -> None:
        """Connect to all hosts defined in the host file."""
        if self.is_exiting:
            return

        if not self.hosts:
            self.add_output(
                [
                    (
                        "status_error",
                        f"No hosts found in '{self.host_file}'. Check file and tags.",
                    )
                ]
            )
            return

        connect_tasks = [
            asyncio.create_task(self.connect_host(*host_details))
            for host_details in self.hosts
        ]
        for task in connect_tasks:
            self.async_tasks.add(task)
            task.add_done_callback(self.async_tasks.discard)

        await asyncio.gather(*connect_tasks, return_exceptions=True)

        if self.initial_command and not self.is_exiting:
            self.input_field.set_edit_text(self.initial_command)
            self.process_command(self.initial_command)

    def process_command(self, command: str) -> None:
        """Process a command entered in the input field."""
        if self.is_exiting or not command.strip():
            return

        command = command.strip()
        if command.lower() == "exit":
            self.initiate_exit()
            return

        self.add_output([("command_echo", f">>> {command}")])
        self.input_field.set_edit_text("")

        for host_name, conn in self.connections.items():
            if self.is_exiting:
                break
            if conn and not conn.is_closed():
                task = asyncio.create_task(
                    self.run_command_on_host(host_name, conn, command)
                )
                self.async_tasks.add(task)
                task.add_done_callback(self.async_tasks.discard)
            else:
                prompt = format_host_prompt(host_name, self.max_name_length)
                self.add_output(
                    prompt + [("status_error", "Not connected, skipping.")]
                )

    async def run_command_on_host(
        self,
        host_name: str,
        conn: asyncssh.SSHClientConnection,
        command: str,
    ) -> None:
        """Run a command on a specific host and stream the output."""
        if self.is_exiting:  # If exiting, do not run commands
            return

        prompt = format_host_prompt(host_name, self.max_name_length)

        cols = 80
        if self.loop and self.loop.screen:
            cols = self.loop.screen.get_cols_rows()[0]
        remote_width = (
            max(cols - self.max_name_length - 3, 10) - 1
        )  # Decrease 1 column for the scroollbar.

        output_queue: asyncio.Queue[str | None] = self.output_queues[host_name]

        stream_task = asyncio.create_task(
            stream_command_output(
                conn, command, remote_width, output_queue, color=True
            )
        )
        self.async_tasks.add(stream_task)
        stream_task.add_done_callback(self.async_tasks.discard)

        try:
            if self.separate_output:
                collected_output: List[str] = []
                while not self.is_exiting:
                    try:
                        line_data = await asyncio.wait_for(
                            output_queue.get(), timeout=0.1
                        )
                    except asyncio.TimeoutError:
                        if stream_task.done():
                            break
                        continue
                    if line_data is None:
                        break
                    collected_output.append(line_data)
                for line_data in collected_output:
                    processed_line_markup = ansi_to_urwid_markup(
                        line_data.rstrip("\r\n")
                    )
                    if processed_line_markup:
                        self.add_output(prompt + processed_line_markup)
                    elif self.allow_empty_line and line_data.strip() == "":
                        self.add_output(prompt + [""])
            else:
                while not self.is_exiting:
                    try:
                        line_data = await asyncio.wait_for(
                            output_queue.get(), timeout=0.1
                        )
                    except asyncio.TimeoutError:
                        if stream_task.done():
                            break
                        continue

                    if line_data is None:
                        break

                    processed_line_markup = ansi_to_urwid_markup(
                        line_data.rstrip("\r\n")
                    )
                    if processed_line_markup:
                        self.add_output(prompt + processed_line_markup)
                    elif (
                        self.allow_empty_line
                        and line_data.strip() == ""
                        and not processed_line_markup
                    ):
                        self.add_output(prompt + [""])

        except Exception as e:
            if not self.is_exiting:
                self.add_output(
                    prompt
                    + [("status_error", f"Cmd error: {type(e).__name__} {e}")]
                )
        finally:
            if not stream_task.done():
                stream_task.cancel()
            try:
                await stream_task
            except asyncio.CancelledError:
                if not self.is_exiting:
                    self.add_output(
                        prompt
                        + [("status_neutral", "Command cancelled/interrupted.")]
                    )
            except Exception:
                pass

    def update_prompt_attribute(self, *args, **kwargs):
        """Update input prompt color"""
        if self.main_pile.focus_position == 2:
            self.prompt_attr_map.set_attr_map({None: "input_prompt"})
        else:
            self.prompt_attr_map.set_attr_map({None: "input_prompt_inactive"})
        if self.loop and self.loop.event_loop and not self.draw_screen_handle:
            self.draw_screen_handle = self.loop.event_loop.alarm(
                0, self._request_draw
            )

    def handle_input(self, key: str) -> bool | None:
        """Handle user input from the keyboard."""
        if self.is_exiting:
            return True

        self.update_prompt_attribute()

        if key == "enter":
            self.process_command(self.input_field.edit_text)
            return True
        if key in ("ctrl d", "ctrl c"):
            self.initiate_exit()
            return True

        return None

    def initiate_exit(self) -> None:
        """Initiate the exit process for the TUI."""
        if self.is_exiting:
            return
        self.is_exiting = True

        if self.asyncio_loop and not self.asyncio_loop.is_closed():
            self.shutdown_task = self.asyncio_loop.create_task(
                self.perform_shutdown()
            )
        else:
            self._direct_exit_loop()

    def _direct_exit_loop(self) -> None:
        """Directly exit the Urwid main loop without waiting for async tasks."""
        if self.loop:
            raise urwid.ExitMainLoop()

    async def perform_shutdown(self) -> None:
        """Perform the shutdown process for the TUI."""
        self.add_output(
            [("status_neutral", "Exiting... Closing connections...")]
        )

        close_conn_tasks = []
        for host_name, conn in self.connections.items():
            if conn and not conn.is_closed():
                self.add_output(
                    format_host_prompt(host_name, self.max_name_length)
                    + [("status_neutral", "Closing...")]
                )
                close_conn_tasks.append(
                    asyncio.create_task(self._close_single_connection(conn))
                )

        if close_conn_tasks:
            await asyncio.gather(*close_conn_tasks, return_exceptions=True)
        self.add_output(
            [("status_neutral", "All connections closed or timed out.")]
        )

        if self.async_tasks:
            self.add_output(
                [
                    (
                        "status_neutral",
                        f"Cleaning up {len(self.async_tasks)} tasks...",
                    )
                ]
            )
            for task in list(self.async_tasks):
                if not task.done():
                    task.cancel()
            await asyncio.gather(*self.async_tasks, return_exceptions=True)
            self.async_tasks.clear()

        self.add_output(
            [("status_neutral", "Cleanup complete. Ananta TUI will now exit.")]
        )

        if self.loop and self.loop.event_loop:
            self.loop.event_loop.alarm(0, self._direct_exit_loop)

    async def _close_single_connection(
        self,
        conn: asyncssh.SSHClientConnection,
    ) -> None:
        """Close a single SSH connection gracefully."""
        try:
            conn.close()
            await asyncio.wait_for(conn.wait_closed(), timeout=2.0)
        except (asyncio.TimeoutError, Exception):
            pass

    def _initial_setup_tasks(self, *_args: Any) -> None:
        """Perform initial setup tasks after the main loop starts."""
        if self.asyncio_loop and not self.asyncio_loop.is_closed():
            self.asyncio_loop.create_task(self.connect_all_hosts())
        else:
            self.add_output(
                [
                    (
                        "status_error",
                        "Asyncio loop not available for initial tasks.",
                    )
                ]
            )

    def run(self) -> None:
        """Run the Ananta TUI main loop."""
        try:
            self.asyncio_loop = asyncio.get_event_loop_policy().get_event_loop()
        except RuntimeError:
            self.asyncio_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.asyncio_loop)

        urwid_event_loop = urwid.AsyncioEventLoop(loop=self.asyncio_loop)

        self.loop = AnantaMainLoop(
            widget=self.main_layout,
            palette=self.current_palette,  # Use the pre-built palette
            event_loop=urwid_event_loop,
            unhandled_input=self.handle_input,
        )

        self.main_pile.focus_position = 2
        if isinstance(self.input_wrapper, urwid.Columns):
            self.input_wrapper.focus_position = 1
        elif hasattr(self.input_wrapper, "focus_col"):
            self.input_wrapper.focus_col = 1

        try:
            self.loop.screen.set_terminal_properties(colors=256)
            self.loop.screen.set_mouse_tracking(True)
        except Exception:
            pass

        self.loop.event_loop.alarm(0, self._initial_setup_tasks)

        try:
            self.loop.run()
        except urwid.ExitMainLoop:
            print("\nAnanta TUI exiting normally.")
        except KeyboardInterrupt:
            print("\nAnanta TUI interrupted by user (KeyboardInterrupt).")
            if not self.is_exiting:
                self.initiate_exit()

        except Exception as e:
            print(f"\nAnanta TUI encountered an unexpected error: {e}")
            import traceback

            traceback.print_exc()
        finally:
            if not self.is_exiting:
                self.is_exiting = True

            if self.asyncio_loop and not self.asyncio_loop.is_closed():
                try:
                    pending = asyncio.all_tasks(self.asyncio_loop)
                    if pending:
                        self.asyncio_loop.run_until_complete(
                            asyncio.gather(*pending, return_exceptions=True)
                        )
                except RuntimeError:
                    pass
                finally:
                    if (
                        not self.asyncio_loop.is_closed()
                    ):  # Check again before closing
                        self.asyncio_loop.close()
            print("Ananta TUI has finished.")



#!/usr/bin/env python
"""
Ananta: a command-line tool that allows users to execute commands on multiple
remote hosts at once via SSH. With Ananta, you can streamline your workflow,
automate repetitive tasks, and save time and effort.
"""

from . import __version__

from .config import get_hosts
from .output import print_output  # Used by non-TUI mode
from .ssh import execute  # Used by non-TUI mode
from types import ModuleType
from typing import Dict, List
import argparse
import asyncio
import os
import sys

uvloop: ModuleType | None = None
try:
    if sys.platform == "win32":
        import winloop as uvloop
    else:
        import uvloop
except ImportError:
    pass  # uvloop or winloop is an optional for speedup, not a requirement


async def main(  # This is the non-TUI main function
    host_file: str,
    ssh_command: str,
    local_display_width: int,
    separate_output: bool,
    allow_empty_line: bool,
    allow_cursor_control: bool,
    default_key: str | None,
    color: bool,
    host_tags: str | None,
) -> None:
    """Main function to execute commands on multiple remote hosts (non-TUI mode)."""

    hosts_to_execute, max_name_length = get_hosts(host_file, host_tags)

    if not hosts_to_execute:
        print("No hosts found to execute the command on.")
        return

    # Dictionary to hold separate output queues for each host
    output_queues: Dict[str, asyncio.Queue[str | None]] = {
        host_name: asyncio.Queue() for host_name, *_ in hosts_to_execute
    }

    # Create a lock for synchronizing output printing
    print_lock = asyncio.Lock()

    # Create a separate task for each host to print the output
    print_tasks = [
        print_output(
            host_name,
            max_name_length,
            allow_empty_line,
            allow_cursor_control,
            separate_output,
            print_lock,
            output_queues[host_name],
            color,
        )
        for host_name, *_ in hosts_to_execute
    ]
    # Ensure future is created for Python 3.10 compatibility with some asyncio versions
    # For modern asyncio, ensure_future is often not needed for create_task results.
    # However, gathering them is the main point.
    # asyncio.gather itself will schedule them if they are bare coroutines.
    # If they are already tasks, gather works fine.
    # Let's keep ensure_future for clarity or remove if targeting only very new Python/asyncio.
    # For now, let's rely on gather to handle it.
    # asyncio.ensure_future(asyncio.gather(*print_tasks)) # Old way
    # Modern way: just gather the coroutines/tasks.
    # If print_output are coroutines, gather will schedule them.
    # If they are already tasks (e.g. from create_task), gather awaits them.
    # The list comprehension creates coroutines, so gather will schedule them.

    # We need to ensure print_tasks are running in the background
    # while execute tasks are also running.
    # One way is to create tasks for them and then gather execute tasks.

    printing_task_group = asyncio.gather(*print_tasks)

    # Create a task for each host to execute the SSH command
    exec_tasks = [
        execute(
            host_name,
            ip_address,
            ssh_port,
            username,
            key_path,
            ssh_command,
            max_name_length,
            local_display_width,
            separate_output,
            default_key,
            output_queues[host_name],  # Pass the queue here
            color,
        )
        for host_name, ip_address, ssh_port, username, key_path in hosts_to_execute
    ]

    # Execute all command execution tasks concurrently
    await asyncio.gather(*exec_tasks)

    # After all exec_tasks are done, signal end to print_tasks
    for host_name in output_queues:
        await output_queues[host_name].put(None)

    # Wait for all printing tasks to complete
    await printing_task_group


def run_cli() -> None:
    """Command-line interface for Ananta."""
    parser = argparse.ArgumentParser(
        description="Execute commands on multiple remote hosts via SSH."
    )
    parser.add_argument(
        "host_file",
        nargs="?",
        default=None,
        help="File containing host information",
    )
    parser.add_argument(
        "command",
        nargs=argparse.REMAINDER,
        help="Command to execute on remote hosts",
    )
    parser.add_argument(
        "--tui",
        action="store_true",
        help="Launch the Urwid-based Text User Interface (TUI) mode.",  # Updated help text
    )
    parser.add_argument(
        "-n",
        "-N",
        "--no-color",
        action="store_true",
        help="Disable host coloring (for non-TUI mode)",  # Clarified help
    )
    parser.add_argument(
        "-s",
        "-S",
        "--separate-output",
        action="store_true",
        help="Print output from each host without interleaving (for non-TUI mode)",  # Clarified help
    )
    parser.add_argument(
        "-t",
        "-T",
        "--host-tags",
        type=str,
        help="Host's tag(s) (comma separated)",
    )
    parser.add_argument(
        "-w",
        "-W",
        "--terminal-width",
        type=int,
        help="Set terminal width (for non-TUI mode)",  # Clarified help
    )
    parser.add_argument(
        "-e",
        "-E",
        "--allow-empty-line",
        action="store_true",
        help="Allow printing the empty line (for non-TUI mode)",  # Clarified help
    )
    parser.add_argument(
        "-c",
        "-C",
        "--allow-cursor-control",
        action="store_true",
        help=(
            "Allow cursor control codes (for non-TUI mode; "
            "useful for commands like fastfetch or neofetch)"
        ),  # Clarified help
    )
    parser.add_argument(
        "-v",
        "-V",
        "--version",
        action="store_true",
        help="Show the version of Ananta",
    )
    parser.add_argument(
        "-k",
        "-K",
        "--default-key",
        type=str,
        help="Path to default SSH private key",
    )
    args: argparse.Namespace = parser.parse_args()

    if uvloop:
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    if args.version:
        # Print the version of Ananta with the asyncio event loop module
        print(
            f"Ananta-{__version__} "
            f"powered by {asyncio.get_event_loop_policy().__module__}"
        )
        sys.exit(0)

    host_file: str | None = args.host_file
    ssh_command_list: List[str] = (
        args.command
    )  # Keep as list for TUI initial command
    ssh_command_str: str = " ".join(ssh_command_list)

    if not host_file:
        parser.print_help()
        sys.exit(0)

    if args.tui:
        try:
            import urwid  # noqa -- check if urwid is installed
        except ImportError:
            print(
                "Error: 'urwid' library is required for TUI mode but is not installed."
            )
            print("Please install it, for example: pip install urwid")
            sys.exit(1)

        # Assuming the new Urwid TUI class is in ananta.tui
        from ananta.tui import AnantaUrwidTUI

        app = AnantaUrwidTUI(
            host_file=host_file,  # Already checked it's not None
            initial_command=(
                ssh_command_str if ssh_command_str.strip() else None
            ),
            host_tags=args.host_tags,
            default_key=args.default_key,
            separate_output=args.separate_output,
            allow_empty_line=args.allow_empty_line,
        )
        app.run()  # This will block until the TUI exits
        sys.exit(0)  # Exit after TUI finishes

    # Non-TUI mode continues from here
    if (
        not ssh_command_str.strip()
    ):  # Check if command is empty for non-TUI mode
        print("Error: No command specified for non-TUI mode.")
        parser.print_help()
        sys.exit(1)  # Exit with error if no command for non-TUI

    try:
        local_display_width: int = args.terminal_width or int(
            os.environ.get("COLUMNS", os.get_terminal_size().columns)
        )
    except OSError:
        local_display_width = args.terminal_width or 80

    color = not args.no_color

    asyncio.run(
        main(
            host_file,
            ssh_command_str,
            local_display_width,
            args.separate_output,
            args.allow_empty_line,
            args.allow_cursor_control,
            args.default_key,
            color,
            args.host_tags,
        )
    )


if __name__ == "__main__":
    run_cli()



import re
import urwid
from typing import List, Tuple
from dataclasses import dataclass, field


_DEFAULT_FG_COLOR = "default"  # Urwid's default color string for foreground
_DEFAULT_BG_COLOR = "default"  # Urwid's default color string for background


@dataclass
class _AnsiState:
    """Manages the current state of ANSI SGR attributes."""

    """Initialize the ANSI state with default attributes."""
    fg: str = _DEFAULT_FG_COLOR
    bg: str = _DEFAULT_BG_COLOR
    styles: set[str] = field(default_factory=set)

    def reset(self):
        """Reset all attributes to default."""
        self.fg = _DEFAULT_FG_COLOR
        self.bg = _DEFAULT_BG_COLOR
        self.styles.clear()

    def get_attr_spec(self) -> urwid.AttrSpec:
        """Create an Urwid AttrSpec from the current state."""
        current_fg_color = self.fg
        current_bg_color = self.bg

        # Styles that Urwid's AttrSpec parses from the foreground string
        urwid_supported_styles = {
            "bold",
            "underline",
            "standout",
            "italics",
            "blink",
            "strikethrough",
        }

        active_style_parts = [
            s for s in self.styles if s in urwid_supported_styles
        ]

        # Handle 'reverse' by swapping fg and bg colors
        if "reverse" in self.styles:
            current_fg_color, current_bg_color = (
                current_bg_color,
                current_fg_color,
            )

        # Handle 'conceal' by making fg the same as bg
        if "conceal" in self.styles:
            current_fg_color = current_bg_color

        # Construct the foreground specification string
        fg_spec_parts = []
        if active_style_parts:
            fg_spec_parts.extend(sorted(active_style_parts))

        if current_fg_color != _DEFAULT_FG_COLOR or not fg_spec_parts:
            fg_spec_parts.append(current_fg_color)

        final_fg_spec = ",".join(fg_spec_parts)
        if not final_fg_spec:
            final_fg_spec = _DEFAULT_FG_COLOR

        final_bg_spec = current_bg_color

        return urwid.AttrSpec(final_fg_spec, final_bg_spec)


_ANSI_CONTROL_SEQUENCES = re.compile(
    r"[\x00-\x08\x0B\x0C\x0E-\x1A\x1C-\x1F]"  # C0 control characters
    r"[\x80-\x9F]"  # C1 control characters
    r"\x1bP[^\x1b]*\x1b\\"  # DCS (Device Control String)
    r"\x1b_[^\x1b]*\x1b\\"  # APC (Application Program Command)
    r"\x1b\^[^\x1b]*\x1b\\"  # PM (Privacy Message)
)
_OSC_CONTROL_SEQUENCES = re.compile(
    r"\x1b\][^\x07\x1b]*(\x07|\x1b\\)", re.DOTALL
)
_CSI_CONTROL_SEQUENCES = re.compile(r"(\x1b\[[0-9;?]*)([A-Za-z])")


def _strip_ansi_control_sequences(text: str) -> str:
    """
    Strips non-SGR ANSI escape sequences and other problematic control characters.
    Tabs are NOT handled here; they are expanded later on plain text segments.
    Importantly, \x1b (ESC) is NOT stripped by this function if it's part of an SGR.
    """
    text = _ANSI_CONTROL_SEQUENCES.sub("", text)
    text = _OSC_CONTROL_SEQUENCES.sub("", text)

    # Strip CSI (Control Sequence Introducer) sequences that are NOT SGR (ending in 'm')
    def csi_stripper(match: re.Match) -> str:
        if match.group(2) != "m":
            return ""
        return match.group(0)

    text = _CSI_CONTROL_SEQUENCES.sub(csi_stripper, text)

    if "\r" in text:
        if not text.endswith("\r") and not text.endswith("\r\n"):
            text = text.split("\r")[-1]

    return text


def _expand_tabs_with_col_tracking(
    text: str, starting_col: int, tab_width: int = 8  # default tab width
) -> tuple[str, int]:
    """Expands tabs in a string to spaces, tracking column position."""
    expanded_text = []
    current_col = starting_col
    for char in text:
        if char == "\t":
            spaces_to_add = tab_width - (current_col % tab_width)
            if spaces_to_add == 0 and tab_width > 0:
                spaces_to_add = tab_width
            expanded_text.append(" " * spaces_to_add)
            current_col += spaces_to_add
        else:
            expanded_text.append(char)
            current_col += 1
    return "".join(expanded_text), current_col


def _handle_extended_color(
    params: list[str], idx: int, state: _AnsiState, is_fg: bool
) -> int:
    """Handle extended color codes (38;5;n or 38;2;r;g;b for fg, 48 for bg)."""
    if idx >= len(params):
        return idx
    color_mode = params[idx]
    idx += 1
    if color_mode == "5":
        if idx < len(params):
            color_id = params[idx]
            try:
                int(color_id)
                color_val = f"h{color_id}"
                if is_fg:
                    state.fg = color_val
                else:
                    state.bg = color_val
            except ValueError:
                pass
            idx += 1
    elif color_mode == "2":
        if idx + 2 < len(params):
            try:
                r, g, b = params[idx : idx + 3]
                color_val = f"#{int(r):02x}{int(g):02x}{int(b):02x}"
                if is_fg:
                    state.fg = color_val
                else:
                    state.bg = color_val
            except ValueError:
                pass
            idx += 3
        else:
            # Malformed sequence, consume rest of params to avoid misinterpretation.
            return len(params)
    return idx


_ANSI_SGR_PATTERN = re.compile(r"\x1b\[([\d;]*)m")
_ANSI_FG_COLOR_MAP = {
    "30": "black",
    "31": "dark red",
    "32": "dark green",
    "33": "brown",
    "34": "dark blue",
    "35": "dark magenta",
    "36": "dark cyan",
    "37": "light gray",
    "90": "dark gray",
    "91": "light red",
    "92": "light green",
    "93": "yellow",
    "94": "light blue",
    "95": "light magenta",
    "96": "light cyan",
    "97": "white",
}
_ANSI_BG_COLOR_MAP = {
    "40": "black",
    "41": "dark red",
    "42": "dark green",
    "43": "brown",
    "44": "dark blue",
    "45": "dark magenta",
    "46": "dark cyan",
    "47": "light gray",
    "100": "dark gray",
    "101": "light red",
    "102": "light green",
    "103": "yellow",
    "104": "light blue",
    "105": "light magenta",
    "106": "light cyan",
    "107": "white",
}


def ansi_to_urwid_markup(line: str) -> List[Tuple[urwid.AttrSpec, str] | str]:
    """
    Convert a string containing ANSI SGR codes to Urwid markup list.
    Non-SGR control codes are stripped, and tabs are expanded.
    """
    cleaned_line = _strip_ansi_control_sequences(line)

    markup: List[Tuple[urwid.AttrSpec, str] | str] = []
    last_pos = 0
    current_col = 0
    state = _AnsiState()

    for match in _ANSI_SGR_PATTERN.finditer(cleaned_line):
        start, end = match.span()

        if start > last_pos:
            text_segment = cleaned_line[last_pos:start]
            expanded_segment, new_col = _expand_tabs_with_col_tracking(
                text_segment, current_col
            )
            current_col = new_col
            if expanded_segment:
                markup.append((state.get_attr_spec(), expanded_segment))

        last_pos = end

        codes_str = match.group(1)
        if not codes_str or codes_str == "0":
            state.reset()
        else:
            params = codes_str.split(";")
            idx = 0
            while idx < len(params):
                code = params[idx]
                if not code:
                    idx += 1
                    continue
                # turn off black formatting to make code more readable
                # fmt: off
                match code:
                    # add, update, or remove styles based on ANSI codes 
                    case "1": state.styles.add("bold")
                    case "2": state.styles.add("faint")
                    case "3": state.styles.add("italics")
                    case "4": state.styles.add("underline")
                    case "5" | "6": state.styles.add("blink")
                    case "7": state.styles.update({"reverse", "standout"})
                    case "8": state.styles.add("conceal")
                    case "9": state.styles.add("strikethrough")
                    case "21": state.styles.add("underline")
                    case "22": state.styles.difference_update({"bold", "faint"})
                    case "23": state.styles.discard("italics")
                    case "24": state.styles.discard("underline")
                    case "25": state.styles.discard("blink")
                    case "27": state.styles.difference_update({"reverse", "standout"})
                    case "28": state.styles.discard("conceal")
                    case "29": state.styles.discard("strikethrough")
                    # foreground and background color codes
                    case code if code in _ANSI_FG_COLOR_MAP:
                        state.fg = _ANSI_FG_COLOR_MAP[code]
                    case "39":
                        state.fg = _DEFAULT_FG_COLOR
                    case code if code in _ANSI_BG_COLOR_MAP:
                        state.bg = _ANSI_BG_COLOR_MAP[code]
                    case "49":
                        state.bg = _DEFAULT_BG_COLOR
                    # extended color codes
                    case "38":
                        idx = _handle_extended_color(
                            params, idx + 1, state, True
                        )
                        continue
                    case "48":
                        idx = _handle_extended_color(
                            params, idx + 1, state, False
                        )
                        continue
                    # do nothing for unsupported or unrecognized codes
                    case _:
                        pass
                # fmt: on
                # turn on black formatting
                idx += 1

    if last_pos < len(cleaned_line):
        text_segment = cleaned_line[last_pos:]
        expanded_segment, _ = _expand_tabs_with_col_tracking(
            text_segment, current_col
        )
        if expanded_segment:
            markup.append((state.get_attr_spec(), expanded_segment))

    return [
        part
        for part in markup
        if isinstance(part, str) or (isinstance(part, tuple) and part[1])
    ]
